#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <dirent.h> //LikeU 

#include "bmsparser.h"

#define BMS_HEADER 0
#define BMS_DATA 1
#define BMS_ALL 2

#define INVALID -1

#define MAX_LENGTH 2048

#define INDEX_SIZE 5
#define WAV_INDEX 3
#define BMP_INDEX 3


static char *header_idx[HEADER_NUM]={"TITLE", "GENRE", "ARTIST",
    "BPM", "PLAYER", "PLAYLEVEL", "RANK", "TOTAL"};

void DumpNotes(BMS_Notes *notes); 
BMS_Notes *BubbleSortNotes(BMS_Notes *notes); 

int AdjustFilename(char *dir_name, char *file_name); // LikeU  

BMS * CreateBMS(char *filename) {
    BMS_Header *header=NULL;
    BMS_Data *data=NULL;
    BMS_Data_Info *info=NULL;
    BMS *bms=NULL;

    int i=0;

    /* ʱ ޸  */
    bms = (BMS *) malloc (sizeof(BMS));
    if (bms == NULL) {
        fprintf(stderr, "BMS Memory allocation failed\n");
        return NULL;
    }

    header=(BMS_Header *) malloc (sizeof(BMS_Header));
    if (bms == NULL) {
        fprintf(stderr, "BMS Header Memory allocation failed\n");
        return NULL;
    }

    data=(BMS_Data *) malloc (sizeof(BMS_Data));
    if(data==NULL) {
        fprintf(stderr, "BMS Data Memory allocation failed\n");
        return NULL;
    }
    info =(BMS_Data_Info *) malloc (sizeof(BMS_Data_Info));
    if(data==NULL) {
        fprintf(stderr, "BMS Data Information Memory allocation failed\n");
        return NULL;
    }

    GetFilenameAndPath(header, filename);

#ifdef DEBUG
    fprintf(stdout, "path : %s\n", header->path);
    fprintf(stdout, "path : %s\n", header->filename);
#endif

    /*  ̸  */
    if (bms->header->filename == NULL) {
        fprintf(stderr, "File doesn't exist !\n");
        return NULL;
    }

    bms->header=header;
    bms->data=data;
    bms->data->info=info;

    /*BMS Header ʱȭ */
    bms->header->title=NULL;
    bms->header->genre=NULL;
    bms->header->artist=NULL;
    bms->header->bpm=-1;
    bms->header->player=-1;
    bms->header->playlevel=-1;
    bms->header->rank=-1;
    bms->header->total=-1;

    /* BMS Data  ʱȭ */
    bms->data->info->max_measure=0;
    bms->data->info->max_key=0;
    bms->data->info->total_st_measure=0;
    bms->data->info->total_notes=0;
    bms->data->info->total_bgm=0;
    bms->data->info->total_bga=0;
    bms->data->info->total_bpm=0;
    bms->data->info->measure_length=0;
    bms->data->info->bga_flag=BGA_NO_EXIST;

    bms->data->notes=NULL;
    bms->data->bgm=NULL;
    bms->data->bga=NULL;
    bms->data->bpm=NULL;
    bms->data->last_note=NULL;
    bms->data->last_bgm=NULL;
    bms->data->m_ratio=NULL;
    bms->data->last_m_ratio=NULL;

    for (i=0 ; i < MAX_WAV ; i++) {
        bms->data->wav[i].index=-1;
        bms->data->wav[i].filename=NULL;
        bms->data->wav[i].mix=NULL;
    }
    for (i=0 ; i < MAX_BMP ; i++) {
        bms->data->bmp[i].index=-1;
        bms->data->bmp[i].filename=NULL;
        bms->data->bmp[i].scr=NULL;
    }
    return bms;
}

void DestroyBMS(BMS *bms) {

    int i=0;
    BMS_Notes *temp;
    BMS_Notes *temp_next;

    if (bms!=NULL) {
        if (bms->header != NULL) {
            if (!(bms->header->path==NULL)) {
                free(bms->header->path); /* ϰ ڿ ޸  */
            }
            if (!(bms->header->filename==NULL)) {
                free(bms->header->filename); /* ̸ ڿ ޸  */
            }
            if (!(bms->header->title==NULL)) {
                free(bms->header->title);	/* ŸƲ ڿ ޸  */
            }
            if (!(bms->header->genre==NULL)) {
                free(bms->header->genre);	/* 帣 ڿ ޸  */
            }
            if (!(bms->header->artist==NULL)) {
                free(bms->header->artist); /* ƼƮ ڿ ޸  */
            }
            free(bms->header);
        }

        /* BMS Info ü ޸  */
        if (bms->data->info != NULL) {
            free(bms->data->info);
        }

        /* BMP, BGA ޸  */
        for (i=0; i < MAX_WAV ; i++) {
            if (bms->data->wav[i].mix != NULL) {
                Mix_FreeChunk(bms->data->wav[i].mix);
            }
        }
        for (i=0; i < MAX_BMP ; i++) {
            if (bms->data->bmp[i].scr != NULL) {
                SDL_FreeSurface(bms->data->bmp[i].scr);
            }
        }

        /* BMS notes ũ Ʈ ޸  */
	temp = bms->data->notes;
        while (temp != NULL) {
	    temp_next = temp->next;
            free(temp);
	    temp = temp_next;
        }
	temp = bms->data->bgm;
        while (temp != NULL) {
	    temp_next = temp->next;
            free(temp);
	    temp = temp_next;
        }
	temp = bms->data->bga;
        while (temp != NULL) {
	    temp_next = temp->next;
            free(temp);
	    temp = temp_next;
        }
        free(bms);
    }

}

int GetBMS(char *file ,int flag, BMS *bms) {

    FILE *fp=NULL;
    char *str=NULL;
    int pos=0; /*  ڿ ġ */


    str=(char *) malloc (MAX_LENGTH);
    if (str== NULL ) {
        fprintf(stderr, "Memory allocatino failed\n");
        return 0;
    }
    memset(str, 0, MAX_LENGTH); /* ޸ ʱȭ */

    fp=fopen(file, "r");
    if (fp == NULL) {
        fprintf(stderr, "Error opening file : %s\n", file);
        return 0;
    }

    if (flag == BMS_DATA) {
        /* Ʈ ޸ Ҵ */
	printf ("bms->data->notes = %p\n", bms->data->notes);
		
#if 0
        if (bms->data->info->total_notes !=0) {
            bms->data->notes=(BMS_Notes *) calloc (sizeof(BMS_Notes), bms->data->info->total_notes+1);
        }
        bms->data->last_note=bms->data->notes;

        /* bgm ޸ Ҵ */
        if (bms->data->info->total_bgm != 0 ) {
            bms->data->bgm=(BMS_Notes *) calloc (sizeof(BMS_Notes), bms->data->info->total_bgm+1);
        }
        bms->data->last_bgm=bms->data->bgm;

        /* bga ޸ Ҵ */
        if (bms->data->info->total_bga != 0 ) {
            bms->data->bga=(BMS_Notes *) calloc (sizeof(BMS_Notes), bms->data->info->total_bga+1);
        }
        bms->data->last_bga=bms->data->bga;

	/* bpm ޸ Ҵ */
        if (bms->data->info->total_bpm != 0 || 1) {
            bms->data->bpm=(BMS_Notes *) calloc (sizeof(BMS_Notes), bms->data->info->total_bpm+1);
        }
        bms->data->last_bpm=bms->data->bpm;
#endif
        /* measure ratio   ޸ Ҵ */
        if (bms->data->info->total_st_measure != 0) {
            bms->data->m_ratio= (Measure *) calloc (sizeof (Measure), bms->data->info->total_st_measure+1);
        }
        bms->data->last_m_ratio=bms->data->m_ratio;

    }

    while(fgets(str, MAX_LENGTH, fp) != NULL) {
        RemoveWhiteSpace(str, FORWARD);
        RemoveWhiteSpace(str, BACKWARD);
        pos=IsValidString(str);
        if (pos == INVALID) {
            continue;
        }
        str+=pos; /* ȿ ڿ  ̵ */

        switch (flag) {
            case BMS_HEADER :
                FillBMSHeader(bms, str);
                break;
            case BMS_DATA:
                FillBMSData(bms, str);
                break;
            default :
                break;
        }
        memset(str, 0, MAX_LENGTH); /* ޸ ʱȭ */
    }

    /*    :
        3  ƴ϶ 3 Ͽ 3  
        4  ƴ϶ 4 Ͽ 4   
       */
    if (flag == BMS_HEADER) {
        if (bms->data->info->measure_length % 4 != 0) {
            bms->data->info->measure_length*=4;
        }
        if (bms->data->info->measure_length % 3 != 0 ) {
            bms->data->info->measure_length*=3;
        }
        bms->data->info->measure_length=192;
    }
    if (flag == BMS_DATA) {
        //QuickSortNotes(bms->data->notes, bms->data->notes, (bms->data->last_note-1));
        bms->data->last_note = BubbleSortNotes(bms->data->notes);
        bms->data->last_bgm = BubbleSortNotes(bms->data->bgm);
	bms->data->last_bga = BubbleSortNotes(bms->data->bga);
	bms->data->last_bpm = BubbleSortNotes(bms->data->bpm);
#ifdef DEBUG
	DumpNotes (bms->data->notes);
	printf ("dump 1\n");
	DumpNotes (bms->data->bgm);
	printf ("dump 2\n");
	DumpNotes (bms->data->bga);
	printf ("dump 3\n");
	DumpNotes (bms->data->bpm);
	printf ("dump 4\n");

#endif
        if (bms->data->last_m_ratio != NULL) {
            ResizeMeasureHeight(bms);
        }
    }

    /* ޸   */
    free(str);
    fclose(fp);
    return 1;
}

int IsValidString(char *str) {
    char *ptr=NULL;
    int count=0;

    ptr=str;

    for (count=0; *ptr!='\0' ; count++, ptr++) {
        if(!isspace(*ptr)){
            break;
        }
    }

    if (*ptr == '#') {  /* #   ȿ ʵ */
        return count;
    }
    else {
        return INVALID;
    }
}

void FillBMSHeader(BMS *bms, char *str) {
    int max_measure=-1;
    int str_count=0;
    int beat=0;
    int max_key=0;
    int measure_length=0;

    char index[INDEX_SIZE];
    char beats_str[3];
    char *beats_pos=NULL;
    char *value=NULL;

    str++; /* #  ġ ̵ */
    RemoveWhiteSpace(str, FORWARD);

    /* Header  Ѵ */
    if ( SetHeader(str, bms) == 1) {
        return;
    }

    /*BMS Extra info  ä */
    memset (index, 0, INDEX_SIZE);
    if((value=GetData(str, index)) != NULL) {
        /* ִ Ű  */
        if (index[3] == '1' || index[3] == '2') {
            max_key= (int) (index[4] - '0');
            if (bms->data->info->max_key < max_key) {
                bms->data->info->max_key= max_key;
            }
        }

        RemoveWhiteSpace (value, FORWARD);
        RemoveWhiteSpace (value, BACKWARD);

        measure_length=GetMeasureLength(value);	/* ȿ ִ Ʈ  Ѵ */
        if (bms->data->info->measure_length < measure_length ) {
            bms->data->info->measure_length=measure_length;
        }

        max_measure = GetMeasureNumber(index);	/* (measure) ȣ Ѵ */
        if (bms->data->info->max_measure < max_measure) {
            bms->data->info->max_measure=max_measure;
        }

        /*measure    ü  Ѵ  */
        if (index[3] == '0' && index[4] == '2') {
            bms->data->info->total_st_measure++;
        }

        beats_pos=value;
        memset(beats_str, '\0', 3);
        while (*beats_pos != '\0')  {
            if (isxdigit(*beats_pos)) {
                beats_str[str_count]=*beats_pos;
                str_count++;
                if (str_count == 2) {
                    str_count=0;
                    beat=(int) strtol(beats_str, (char **) NULL, 16);
                    if (beat != 0) {
                        if (index[3] == '0') {
                            if (index[4] == '1') {
                                bms->data->info->total_bgm++;  /*   Ѵ */
                            }
                            if (index[4] == '4' || index[4] == '6' ||  index[4] == '7') {
                                bms->data->info->total_bga++;	/*BGA   Ѵ */
                            }
                        }
                        if (index[3] == '1' || index[3] == '2') {
                            bms->data->info->total_notes++;
                        }
                        /* for debug */
                        /*
                           fprintf(stdout, "%d\t", beat);
                           */
                    }
                    memset(beats_str, '\0', 3);
                }
            }
            beats_pos++;
        }
    }
}

/* bms ͸  Ѵ */
void FillBMSData(BMS *bms, char *str) {
    FillNotes(bms, str);
    FillWAV(bms, str);
    FillBMP(bms, str);
}

/* bms  Ʈ  о´ */
char * GetData (char *str, char index[]) {
    int index_count=0;
    char *mov_pos=NULL;
    char *value_pos=NULL;

    mov_pos=str;
    if (str != NULL) {
        value_pos=strchr(str, ':');
    }

    if (mov_pos!= NULL && value_pos != NULL) {
        while (mov_pos < value_pos && index_count < 5) {
            if (!isspace(*mov_pos) && isdigit(*mov_pos)) {
                index[index_count]=*mov_pos;
                index_count++;
            }
            mov_pos++;
        }
        if (index_count==5) {		/* BMS  ε 5 10 */
            return value_pos;
        }
        else {
            return NULL;
        }
    }
    return NULL;
}

void DumpNotes(BMS_Notes *notes) {
    BMS_Notes *cur;
    for (cur = notes; cur != NULL; cur=cur->next) {
	printf ("%d %d %d\n", cur->pos, cur->key, cur->data);
    }
}


void FillNotes(BMS *bms, char *str) {

    int str_count=0;
    int note_value=0;
    int current_measure_length=0;	/* beats in measure */
    int measure=-1;
    int measure_height=0;
    int note_count=-1;		/* note   0 */
    int notes_offset=0;
    int key=0;

    char *value=NULL;
    char *notes_pos=NULL;
    char index[INDEX_SIZE];
    char note_str[3];	/* note   ϱ  ӽ ڿ 迭 */


#if 0
    /* ޸𸮰 Ȯ  ٷ  */
    if (bms->data->notes == NULL) {
        return;
    }
#endif
    memset (index, 0, INDEX_SIZE);
    if((value=GetData(str, index)) != NULL) {
        /* :  ̵ */
        RemoveWhiteSpace (value, FORWARD);
        RemoveWhiteSpace (value, BACKWARD);
        if (*value == ':') {
            value++;
        }
        /* ȿ ִ Ʈ  Ѵ */
        current_measure_length=GetMeasureLength(value);	
        /* for debug */
        /*
           fprintf(stdout, "Beats in measure : %d\n", current_measure_length);
           */

        measure=GetMeasureNumber(index);	/* (measure) ȣ Ѵ */
        /* for debug */
        /*
           fprintf(stdout, "measure number : %d\n", measure);
           */

        measure_height=measure*(bms->data->info->measure_length);
        /* measure ratio    ü ִ´ */
        if (index[3] == '0' && index[4] == '2') {
            bms->data->last_m_ratio->pos=measure_height;
            bms->data->last_m_ratio->ratio=atof(value); 
            bms->data->last_m_ratio++;
        }

        /*  мؼ ޸𸮿 ä ִ´*/
        notes_pos=value;

        memset(note_str, '\0', 3);
        while (*notes_pos != '\0')  {
            if (isxdigit(*notes_pos)) {
                note_str[str_count]=*notes_pos;
                str_count++;
                /* 2Ʈ ϳ Ʈ */
                if (str_count == 2) {
                    str_count=0;
                    note_count++;
                    /* ش Ʈ ڵ  ´ */
                    note_value=(int) strtol(note_str, (char **) NULL, 16);
                    if (note_value != 0) {
                        if (index[3] == '0') {
                            /*  */
                            if (index[4] == '1') {
				BMS_Notes *note;

				// alloc
				note = calloc (sizeof(BMS_Notes), 1);
				note->pos = measure_height+notes_offset;
				note->key = 0;  /*bgm  Ű   */
				note->data = note_value;

				/* if bms->data->bgm == NULL, note->next = NULL */
				note->next = bms->data->bgm;
				bms->data->bgm = note;
				
                            }

			    /* 03:  bpm */
                            if (index[4] == '3') {
				BMS_Notes *note;

				// alloc
				note = calloc (sizeof(BMS_Notes), 1);
				note->pos = measure_height+notes_offset;
				note->key = 0;  /*bpm  Ű   */
				note->data = note_value;
				if (bms->data->bpm == NULL) {
				    bms->data->bpm = note;
				    note->next = NULL;
				} else {
				    note->next = bms->data->bpm;
				    bms->data->bpm = note;
				}
				
                            }

                            /* 04, 06, 07:  ̹ */
                            if (bms->data->info->total_bga != 0 ) {
                                if (index[4] == '4' || index[4] == '6' ||  index[4] == '7') {

				    BMS_Notes *note;
                                    key=(int) (index[4] -'0');
                                    notes_offset=(bms->data->info->measure_length/current_measure_length)*note_count;

				    // alloc
				    note = calloc (sizeof(BMS_Notes), 1);
				    note->pos = measure_height+notes_offset;
				    note->key = key;	/*׷ Ÿ key  ü */
				    note->data = note_value;
				    if (bms->data->bga == NULL) {
					bms->data->bga = note;
					note->next = NULL;
				    } else {
					note->next = bms->data->bga;
					bms->data->bga = note;
				    }
                                }
                            }
                        }
                        /* ǹ  1: 1player 2 : 2player*/
                        else if (index[3] == '1' || index[3] == '2') {

			    BMS_Notes *note;

                            key= (int) (index[4] - '0');
                            notes_offset=(bms->data->info->measure_length/current_measure_length)*note_count;

			    // alloc
			    note = calloc (sizeof(BMS_Notes), 1);
			    note->pos = measure_height+notes_offset;
			    note->key =  key;
			    note->data = note_value;

			    if (key ==  0) {
				printf ("key = 0\n");
				abort();
			    }

			    note->next = bms->data->notes;
			    bms->data->notes = note;
				   
			    
                        }
                        /* for debug */
                        /*
                         * 
                         fprintf(stdout, "%d\t", beat);
                         */
                    }
                    memset(note_str, '\0', 3);
                }
            }
            notes_pos++;
        }
    }
}

void FillWAV (BMS *bms, char *str) {

    char fullname[MAX_LENGTH];
    char *wav_str=NULL;
    char wav_index[WAV_INDEX];
    char *filename=NULL;
    char *str_buf=NULL;

    //char check; /* LikeU  */
    
    int index=-1;

    memset (wav_index, '\0', WAV_INDEX);
    memset (fullname, '\0', MAX_LENGTH);

    /* #  ̵ */
    wav_str=str;
    wav_str++;

    /*   */
    RemoveWhiteSpace(wav_str, FORWARD);

    /* ̸ WAV  ڿ ˻ */
    if (!(strncasecmp (wav_str, "WAV", 3))) {
        /* WAV  ŭ ̵ */
        wav_str+=WAV_INDEX;	
        strncpy(wav_index, wav_str, 2);

        /* index  */
        index=(int) strtol(wav_index, (char **)NULL, 16);
        bms->data->wav[index].index=index;

        /* indexŭ ̵ */
        filename=wav_str+3;
        RemoveWhiteSpace(filename , FORWARD);

        /*  ̸  */
        bms->data->wav[index].filename = str_buf =strdup(filename);

	AdjustFilename (bms->header->path,bms->data->wav[index].filename);
        /*    */
	snprintf(fullname, MAX_LENGTH, "%s%c%s", bms->header->path, PATH_SEPERATE, str_buf);
	bms->data->wav[index].mix=Mix_LoadWAV(fullname);

	if (bms->data->wav[index].mix == NULL) {
	    fprintf(stderr, "Can not open wav file : %s (%s)\n", fullname,  Mix_GetError());
	}

#ifdef DEBUG
        fprintf(stdout, "%d\t", bms->data->wav[index].index);
        fprintf(stdout, "%s\n", bms->data->wav[index].filename);
#endif
    }
}

/* Sync filename which ignore upper/lower case.
 * Thanks to LikeU 
 */
int AdjustFilename(char *dir_name, char *file_name){
    struct dirent **namelist;
    int n;
    //char check;
    n=scandir(dir_name,&namelist,0,alphasort);
    while(n--){
        if(strcasecmp(file_name,namelist[n]->d_name)==0){
	    strcpy (file_name, namelist[n]->d_name);
            return 1;
        }
    }
    return 0;
}

void FillBMP (BMS *bms, char *str) {
    char fullname[MAX_LENGTH];
    char *bmp_str=NULL;
    char bmp_index[WAV_INDEX];
    char *filename=NULL;
    char *str_buf=NULL;

    int index=-1;

    memset (bmp_index, '\0', WAV_INDEX);
    memset (fullname, '\0', MAX_LENGTH);

    /* #  ̵ */
    bmp_str=str;
    bmp_str++;

    /*   */
    RemoveWhiteSpace(bmp_str, FORWARD);
    /* ̸ BMP  ڿ ˻ */
    if (!(strncasecmp (bmp_str, "BMP", 3))) {
        /* BMP  ŭ ̵ */
        bmp_str+=BMP_INDEX;	
        strncpy(bmp_index, bmp_str, 2);

        /* index  */
        index=(int) strtol(bmp_index, (char **)NULL, 16);
        bms->data->bmp[index].index=index;

        /* indexŭ ̵ */
        filename=bmp_str+3;
        RemoveWhiteSpace(filename , FORWARD);

        /*  ̸  */
        str_buf = (char *) malloc (strlen(filename)+1);
        memset(str_buf , '\0', strlen(filename)+1);
        strcpy (str_buf , filename);
        bms->data->bmp[index].filename=str_buf;

        /*  ׸   */
        snprintf(fullname, MAX_LENGTH, "%s%c%s", bms->header->path, PATH_SEPERATE, str_buf);
        bms->data->bmp[index].scr=SDL_LoadBMP(fullname);

        if (bms->data->bmp[index].scr == NULL) {
            fprintf(stderr, "Can not open bmp file : %s\n", fullname);
        }

#ifdef DEBUG
        fprintf(stdout, "%d\t", bms->data->bmp[index].index);
        fprintf(stdout, "%s\n", bms->data->bmp[index].filename);
#endif
    }
}

void RemoveWhiteSpace (char *str, int flag) {
    char *last;
    int length=0;
    switch (flag) {
        case BACKWARD :
            length=strlen(str);
            if (length != 0) {
                last=str+length-1;
                while (isspace(*last) && last != str) {
                    last--;
                }
                *(last+1)='\0';
            }
            break;
        case FORWARD :
        default :
            while (isspace(*str) &&  *str != '\0') { /* white space  */
                str++;
            }
            break;
    }
}

int GetMeasureLength(char *value_str) {
    char *temp=NULL;
    int measure_length=0;

    temp=value_str;

    if (value_str == NULL) {
        return 0;
    }

    while (*temp != '\0') {
        if (isxdigit(*temp)) {
            measure_length++;
        }
        temp++;
    }
    measure_length/=2;		/* 2  ̷Ƿ 2  ش */
    return measure_length;
}

int GetMeasureNumber (char index[]) {
    char measure_str[4];	/*  ȣ ϱ  ӽ ڿ 迭 */
    int measure;

    memset(measure_str, '\0', 4);
    memcpy(measure_str, index, 3);

    measure=atoi(measure_str);

    return measure;
}

int SetHeader(char *str, BMS *bms) {
    char *buf=NULL;
    int i_buf=0;
    int i=0;

    for (i=0; i < HEADER_NUM ; i++) {
        if(!strncasecmp(str, header_idx[i], strlen(header_idx[i]))) { /*  index ˻ */
            str+=strlen(header_idx[i]); /* Header Index  value ġ  ̵ */
            RemoveWhiteSpace(str, FORWARD);
            RemoveWhiteSpace(str, BACKWARD);
            switch (i) {
                case TITLE :
                    buf= (char *) malloc (strlen(str)+1);
                    memset(buf, 0, strlen(str)+1);
                    strcpy(buf, str);
                    bms->header->title=buf;
                    break;
                case GENRE :
                    buf= (char *) malloc (strlen(str)+1);
                    memset(buf, 0, strlen(str)+1);
                    strcpy(buf, str);
                    bms->header->genre=buf;
                    break;
                case ARTIST :
                    buf= (char *) malloc (strlen(str)+1);
                    memset(buf, 0, strlen(str)+1);
                    strcpy(buf, str);
                    bms->header->artist=buf;
                case BPM :
                    i_buf=atoi(str);
                    bms->header->bpm=i_buf;
                    break;
                case PLAYER :
                    i_buf=atoi(str);
                    bms->header->player=i_buf;
                    break;
                case PLAYLEVEL :
                    i_buf=atoi(str);
                    bms->header->playlevel=i_buf;
                    break;
                case RANK :
                    i_buf=atoi(str);
                    bms->header->rank=i_buf;
                    break;
                case TOTAL :
                    i_buf=atoi(str);
                    bms->header->total=i_buf;
                    break;
                default :
                    break;
            }
            /* for debug */
            /*
               fprintf(stdout, "%s : %s", header_idx[i], str+strlen(header_idx[i]));
               */
            return 1;
        }
    }
    return 0;
}

int GetFilenameAndPath(BMS_Header * header,  char *filename) {
    char *name_pos=NULL;
    char *path_pos=NULL;
    int str_len=0;
    int path_len=0;

    if (filename == NULL) {
        return -1;
    }

    path_pos=filename;
    str_len=strlen(filename);
    name_pos=strrchr(path_pos, PATH_SEPERATE);
    if(name_pos != NULL) {
        path_len=name_pos -path_pos;
        /* path  */
        header->path = (char *) malloc (path_len+1);
        memset (header->path , '\0', path_len+1);
        strncpy(header->path, path_pos, path_len);

        RemoveWhiteSpace(header->path, FORWARD);
        RemoveWhiteSpace(header->path, BACKWARD);

        /* name  */
        header->filename = (char *) malloc (strlen(name_pos));
        strcpy(header->filename, name_pos+1);
    }
    else {
        header->filename = (char *) malloc (str_len+1);
        strcpy (header->filename , filename);
        return 0;
    }
    RemoveWhiteSpace(header->filename, FORWARD);
    RemoveWhiteSpace(header->filename, BACKWARD);
    return 1;
}

BMS_Notes *BubbleSortNotes(BMS_Notes *notes) {
    BMS_Notes *i, *j;
    BMS_Notes temp;

 
    for (i=notes; i!=NULL; i = i->next) {
	for (j=i->next; j!=NULL; j = j->next) {
	    /*  j ִ  pos i  , j i ٲ۴. */
	    if (j->pos < i->pos) {
		temp.pos = j->pos;
	    	temp.data = j->data;
		temp.key = j->key;

		j->pos = i->pos;
		j->data = i->data;
		j->key = i->key;

		i->pos = temp.pos;
		i->data = temp.data;
		i->key = temp.key;
	    }
	}
    }
    return j;
}

#if 0
void QuickSortNotes(BMS_Notes *notes, BMS_Notes *left, BMS_Notes *right) {
    int pivot;
    BMS_Notes *i=NULL, *j=NULL;

    if ( left < right) {
        i=left;
        j=right+1;
        pivot=left->pos;

        do {
            do {
                i++;
            }
            while (i->pos < pivot);
            do {
                j--;
            }
            while (j->pos > pivot);
            if ( i < j ) {
                SwapNotes (i, j);
            }
        }
        while (i < j);
        SwapNotes (left, j);
        QuickSortNotes(notes, left, j-1);
        QuickSortNotes(notes, j+1, right);

    }
}
#endif

void SwapNotes(BMS_Notes *left, BMS_Notes *right) {
    BMS_Notes temp;

    temp.pos=left->pos;
    temp.key=left->key;
    temp.data=left->data;

    left->pos=right->pos;
    left->key=right->key;
    left->data=right->data;

    right->pos=temp.pos;
    right->key=temp.key;
    right->data=temp.data;
}

void ResizeMeasureHeight (BMS *bms) {
    Measure *current=NULL;
    BMS_Notes *current_notes=NULL;

    int resized_height=0;
    int measure_length=0;

    measure_length=bms->data->info->measure_length;
    current=bms->data->m_ratio;

    while (current != bms->data->last_m_ratio ) {
        current_notes=bms->data->notes;
        resized_height=measure_length*bms->data->m_ratio->ratio;

        /* ġ ˻ؼ ratio    pos  Ѵ. */
        while (current_notes != bms->data->last_note) {
            if (current->pos <= current_notes->pos) {
                current_notes->pos+=resized_height-measure_length;
            }
            current_notes++;
        }
        current++;
    }
}
