#include "define.h"

#ifndef __BMS_PARSER__
#define __BMS_PARSER__

#define BGA_EXIST 1
#define BGA_NO_EXIST 0

#define HEADER_NUM 8

#define BACKWARD 0
#define FORWARD 1

#define BMS_HEADER 0
#define BMS_DATA 1

#define TITLE 0
#define GENRE  1
#define ARTIST 2
#define BPM 3
#define PLAYER 4
#define PLAYLEVEL 5
#define RANK 6
#define TOTAL 7


#define MAX_WAV 256
#define MAX_BMP 256

/* BMS header ü */
typedef struct _BMS_Header {
	char *path;
	char *filename;
	char *title;
	char *genre;
	char *artist;
	int bpm;
	int player;
	int playlevel;
	int rank;
	int total;
}BMS_Header;

/*  ͸  (ְ 256)*/
typedef struct _BMS_WAV {
	int index;	/* ǹ  */
	char *filename;
	Mix_Chunk *mix;
}BMS_WAV;

/*  ͸  (ְ 256)*/
typedef struct _BMS_BMP {
	int index;	
	char *filename;
	SDL_Surface *scr;
}BMS_BMP;

typedef struct _BMS_Notes {
    	struct _BMS_Notes *next;
	int pos;
	int key;
	int data;
}BMS_Notes;

typedef struct _Measure {
	int pos;
	double ratio;
}Measure;

typedef struct _BMS_Data_Info {
	int max_measure;		/* ִ  */
	int max_key;
	int measure_length;		/* Ѹ ִ   */
	int total_st_measure;	/* measure     ü  */
	int total_notes;
	int total_bgm;
	int total_bga;
	int total_bpm;
	int bga_flag;			/*  ׸    */
}BMS_Data_Info;

typedef struct _BMS_Data {
	BMS_Data_Info *info;
	BMS_WAV wav[MAX_WAV];
	BMS_BMP bmp[MAX_BMP];

	/* theses notes are linked list */
	BMS_Notes *notes;
	BMS_Notes *bgm;
	BMS_Notes *bga;
	BMS_Notes *bpm;

	/* theses are linkd list tail */
	BMS_Notes *last_note;
	BMS_Notes *last_bgm;
	BMS_Notes *last_bga;
	BMS_Notes *last_bpm;

	Measure *m_ratio;
	Measure *last_m_ratio;
}BMS_Data;

/* BMS ü */
typedef struct _BMS {
	BMS_Header *header;
	BMS_Data *data;
}BMS;

BMS * CreateBMS(char *filename); /* BMS ü  */

void DestroyBMS(BMS *bms); /* BMS ü ޸  */

int GetBMS (char *, int, BMS *);

int IsValidString(char *);  /* BMS ڿ Ǻ */

void FillBMSHeader(BMS *, char *); /* Header    ´ */

char *GetData (char *, char *); 

void FillBMSData(BMS *, char *); /* Data    ´ */

void FillNotes(BMS *, char *);	/* Note  ä */

void FillWAV (BMS *, char *);	/* WAV  ä */

void FillBMP (BMS *, char *);	/* BMP  ä */

void RemoveWhiteSpace (char *, int ); /*  ڿ    */

int GetMeasureLength(char *);	/*  ̸ Ѵ */

int GetMeasureNumber (char *); /*  ȣ Ѵ */

int SetHeader(char *, BMS *); /* BMS  ä */

int GetFilenameAndPath(BMS_Header *, char *);	/* о ϸ ο ̸ ˻Ѵ. */

void QuickSortNotes(BMS_Notes *notes, BMS_Notes *left, BMS_Notes *right);

void SwapNotes(BMS_Notes *left, BMS_Notes *right);

void ResizeMeasureHeight(BMS *);
#endif
