#include "sdl_pixel.h"
#include "sdl_xft.h"

Uint32 GetMappedPixel(SDL_Xft_WMInfo *sdl_xft, SDL_Surface *surface, 
        unsigned long pixel) {
    Uint32 mpixel;

    SDL_Color color;

    GetSDL_Color(sdl_xft, &color, pixel);
    mpixel=SDL_MapRGB (surface->format, color.r, color.g, color.b);

    return mpixel;
}

void putpixel(SDL_Surface *surface, unsigned int x, unsigned int y, Uint32 pixel) {
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
        case 1:
            *p = pixel;
            break;

        case 2:
            *(Uint16 *)p = pixel;
            break;

        case 3:
            if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
                p[0] = (pixel >> 16) & 0xff;
                p[1] = (pixel >> 8) & 0xff;
                p[2] = pixel & 0xff;
            } else {
                p[0] = pixel & 0xff;
                p[1] = (pixel >> 8) & 0xff;
                p[2] = (pixel >> 16) & 0xff;
            }
            break;
        case 4:
            *(Uint32 *)p = pixel;
            break;
    }
}

void GetSDL_Color(SDL_Xft_WMInfo *sdl_xft, SDL_Color *sdl_color, unsigned long pixel) {
    XColor color;


    /* not tested in big endian */
    color.red=(Uint8) ((pixel & sdl_xft->visual->red_mask) 
            >> sdl_xft->format.r_rshift
            << sdl_xft->format.r_lshift);
    color.green=(Uint8) ((pixel & sdl_xft->visual->green_mask)
            >> sdl_xft->format.g_rshift
            << sdl_xft->format.g_lshift);
    color.blue=(Uint8) ((pixel & sdl_xft->visual->blue_mask) 
            >> sdl_xft->format.b_rshift
            << sdl_xft->format.b_lshift);

    sdl_color->unused=0;
    sdl_color->r=color.red;
    sdl_color->g=color.green;
    sdl_color->b=color.blue;
    /*
       printf("%#lx\t", color.red);
       printf("%#lx\t", color.green);
       printf("%#lx\n", color.blue);
       */
}
