/*
 * gcc -c sdl-xft.c  -I/usr/include/SDL -I/usr/include/X11R6/Xft  -I/usr/include/freetype2/ -L/usr/X11R6/lib -lX11 -lXft -lSDL
 *
 * Changelog :
 *
 * 	04.02.24 SDL_Xft_Get_FontList : Utf-8 support
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sdl_xft.h"
#include "sdl_pixel.h"


/* initialize  sdl_xft */
int SDL_Xft_Init ( SDL_Xft_WMInfo *sdl_xft , SDL_Surface *screen) {

    SDL_SysWMinfo info;

    int bpp;
    int r_mask_size, b_mask_size, g_mask_size;

    /* get window information */
    SDL_VERSION(&info.version);
    if (SDL_GetWMInfo(&info) <0 ) {
        fprintf(stderr , "Error : SDL_GetWMInfo\n");
        return 0;
    }

    /*set display and window infomation */
    sdl_xft->d = info.info.x11.display;
    sdl_xft->w = info.info.x11.window;
    sdl_xft->visual=DefaultVisual(sdl_xft->d,0);

    /* set screen */
    if (screen ==  NULL ) {
        fprintf(stderr, "Error : Can not found primary display\n");
        return 0;
    }
    sdl_xft->screen=screen;

    /* set RGB struct */
    bpp=DefaultDepth(sdl_xft->d , 0);

    r_mask_size= (int) bpp/3;
    g_mask_size= (int) bpp/3 + (int) bpp%3;
    b_mask_size= (int) bpp/3;

    sdl_xft->format.r_rshift = g_mask_size + b_mask_size;
    sdl_xft->format.g_rshift = b_mask_size;
    sdl_xft->format.b_rshift = 0;

    sdl_xft->format.r_lshift = 8- r_mask_size;
    sdl_xft->format.g_lshift = 8- g_mask_size;
    sdl_xft->format.b_lshift = 8- b_mask_size;

    /* set FcConfig */
    sdl_xft->config=FcConfigCreate();
    sdl_xft->config=FcConfigGetCurrent();

    if (sdl_xft->config == NULL) {
        fprintf(stderr, "Error : FcConfig\n");
        return 0;
    }
    return 1;
}

/* get system font set */
FcChar8 **SDL_Xft_Get_FontList (int *fnum, SDL_Xft_WMInfo *sdl_xft) {

    FcFontSet *f_set;
    FcPattern *pattern;

    FcResult result;

    FcChar8 *f_family;
    FcChar8 **fc=NULL;

    int i=0;

    f_set=FcConfigGetFonts(sdl_xft->config, FcSetSystem);;
    *fnum=f_set->nfont;

    if (f_set->nfont == 0) {
        /* fonts don't exist */
        return NULL;
    }

    fc=calloc(f_set->nfont, sizeof (FcChar8 *));

    for (i=0 ; i < f_set->nfont; i++) {
        pattern=f_set->fonts[i];
        result=FcPatternGetString (pattern , FC_FAMILY, 0, &f_family);
        fprintf(stdout, "%s\n", f_family);

        if (result == FcResultMatch) {
            fc[i]=strdup(f_family);
        }
    }
    return fc;

}

void SDL_Xft_FontList_Free (FcChar8 **f_list) {
    free(f_list[0]);
    free(f_list);
}


/* creating new sdl_xft_surface */
SDL_Xft_Surface * SDL_Xft_Create (Uint32 flag, SDL_Xft_WMInfo *sdl_xft) {
    SDL_Xft_Surface *xft_surface;

    Display *d;
    Window w;

    d=sdl_xft->d;
    w=sdl_xft->w;

    /* new memory allocation for SDL_Xft_Surface */
    xft_surface=(SDL_Xft_Surface *) malloc ( sizeof (SDL_Xft_Surface));

    /* set default xftpattern to NULL */
    xft_surface->fcpattern=NULL;

    /* set default value */
    xft_surface->color.colorkey=SDL_XFT_NO_COLORKEY;


    /* set SDL_Surface flag */
    xft_surface->flag=flag;

    /* initialize surface info */
    xft_surface->width=0;
    xft_surface->height=0;

    /* set compact flag */
    xft_surface->compact=DEFAULT;

    /* set default used rectangle */
    xft_surface->used.x=0;
    xft_surface->used.y=0;
    xft_surface->used.w=0;
    xft_surface->used.h=0;
    xft_surface->used.ascent=0;
    xft_surface->used.descent=0;

    /* set default color */

    /* default background color is white */
    xft_surface->color.bg.red = 0xffff;
    xft_surface->color.bg.green = 0xffff;
    xft_surface->color.bg.blue = 0xffff; 
    xft_surface->color.bg.alpha= 0x00ffff; /* alpha doesn't support */

    /* default foregorund color is black */
    xft_surface->color.fg.red = 0x0000;
    xft_surface->color.fg.green = 0x0000;
    xft_surface->color.fg.blue = 0x0000;
    xft_surface->color.fg.alpha= 0x00ffff; /* alpha doesn't not support */

    return xft_surface;
}

/* create pixmap and xftdraw structure */
int SDL_Xft_CreateSurface(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *surface, unsigned int width, unsigned int height) {
    Display *d;
    Window w;

    d=sdl_xft->d;
    w=sdl_xft->w;

    /* create pixmap */
    surface->pixmap=XCreatePixmap(d, w , width, height, DefaultDepth(d,0));

    /*create xftdraw */
    surface->xftdraw=XftDrawCreate(d, surface->pixmap, 
            DefaultVisual (d, 0),
            DefaultColormap (d, 0));

    if (surface->xftdraw == NULL) {
#ifdef DEBUG
        fprintf(stderr, "Can't Create xftdraw structure !\n");
#endif
        return XFTDRAW_ERROR;	/* can't create xftdraw structure */
    }

    /* set pixmap info */
    surface->width=width;
    surface->height=height;
    return 0;

}

void SDL_Xft_Color_Alloc ( SDL_Xft_Color *color , unsigned int flag, 
        Uint16 r, Uint16 g, Uint16 b) {
    switch(flag) {
        case SDL_XFT_BG_COLOR :
            color->bg.red = r;
            color->bg.green = g;
            color->bg.blue = b;
            color->bg.alpha= 0x00ffff; /* alpha doesn't support */
            break;
        case SDL_XFT_FG_COLOR :
            color->fg.red = r;
            color->fg.green = g;
            color->fg.blue = b;
            color->fg.alpha= 0x00ffff; /* alpha doesn't support */
            break;
        default :
            break;
    }
}

/* set colorkey */
void SDL_Xft_Set_Colorkey (SDL_Xft_Color *color , unsigned int flag) {
    color->colorkey=flag;
}

/* FcPatternAddInteger Wrapper */
int SDL_Xft_Pattern_AddInteger (SDL_Xft_Surface *xft_surface, const char *object, int i) {
    return FcPatternAddInteger(xft_surface->fcpattern, object, i);
}

/* FcPatternAddDoubel Wrapper */
int SDL_Xft_Pattern_AddDouble (SDL_Xft_Surface *xft_surface, const char *object, double d) {
    return FcPatternAddDouble (xft_surface->fcpattern, object, d);
}

/* FcPatternAddString Wrapper */
int SDL_Xft_Pattern_AddString (SDL_Xft_Surface *xft_surface, const char *object, const char *s) {
    return FcPatternAddString (xft_surface->fcpattern, object, s);
}

/* FcPatternAddMatrix Wrapper */
int SDL_Xft_Pattern_AddMatrix (SDL_Xft_Surface *xft_surface, const char *object, const FcMatrix *s) {
    return FcPatternAddMatrix (xft_surface->fcpattern, object, s);
}

/* FcPatternAddCharSet Wrapper */
int SDL_Xft_Pattern_AddCharSet (SDL_Xft_Surface *xft_surface, const char *object, const FcCharSet *c) {
    return FcPatternAddCharSet (xft_surface->fcpattern, object, c);
}

/* FcPatternAddBool Wrapper */
int SDL_Xft_Pattern_AddBool (SDL_Xft_Surface *xft_surface, const char *object, FcBool b) {
    return FcPatternAddBool (xft_surface->fcpattern, object, b);
}

/* FcPatternGetInteger Wrapper */
unsigned int SDL_Xft_Pattern_GetInteger(SDL_Xft_Surface *xft_surface, const char *object, int *i) {
    FcResult result;

    result=FcPatternGetInteger(xft_surface->fcpattern, object, 0, i);
    if (result == FcResultTypeMismatch) {
        return FC_PATTERN_GET_FAIL;
    }
    return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetDouble Wrapper */
unsigned int SDL_Xft_Pattern_GetDouble(SDL_Xft_Surface *xft_surface, const char *object, double *d) {
    FcResult result;

    result=FcPatternGetDouble(xft_surface->fcpattern, object, 0, d);
    if (result == FcResultTypeMismatch) {
        return FC_PATTERN_GET_FAIL;
    }
    return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetString Wrapper */
unsigned int SDL_Xft_Pattern_GetString(SDL_Xft_Surface *xft_surface, const char *object, FcChar8 **consts) {
    FcResult result;
    result=FcPatternGetString(xft_surface->fcpattern, object, 0, consts);
    if (result == FcResultTypeMismatch) {
        return FC_PATTERN_GET_FAIL;
    }
    return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetMatrix Wrapper */
unsigned int SDL_Xft_Pattern_GetMatrix(SDL_Xft_Surface *xft_surface, const char *object, FcMatrix **s) {
    FcResult result;

    result=FcPatternGetMatrix(xft_surface->fcpattern, object, 0, s);
    if (result == FcResultTypeMismatch) {
        return FC_PATTERN_GET_FAIL;
    }
    return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetCharSet Wrapper */
unsigned int SDL_Xft_Pattern_GetCharSet(SDL_Xft_Surface *xft_surface, const char *object, FcCharSet **c) {
    FcResult result;

    result=FcPatternGetCharSet(xft_surface->fcpattern, object, 0, c);
    if (result == FcResultTypeMismatch) {
        return FC_PATTERN_GET_FAIL;
    }
    return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetBool Wrapper */
unsigned int SDL_Xft_Pattern_GetBool(SDL_Xft_Surface *xft_surface, const char *object, FcBool *b) {
    FcResult result;

    result=FcPatternGetBool(xft_surface->fcpattern, object, 0, b);
    if (result == FcResultTypeMismatch) {
        return FC_PATTERN_GET_FAIL;
    }
    return FC_PATTERN_GET_SUCCESS;
}

/* Make New FcPattern as FcPatternBuild Wrapper */
void SDL_Xft_Pattern_Build (SDL_Xft_Surface *xft_surface, ...) {

    va_list va;
    FcPattern *set;

    if (xft_surface->fcpattern != NULL) {
        FcPatternDestroy(xft_surface->fcpattern);
    }

    va_start(va, xft_surface);
    set=FcPatternVaBuild(0, va);
    va_end(va);

    xft_surface->fcpattern=FcPatternDuplicate(set);
    FcPatternDestroy(set);
}

void SDL_Xft_Pattern_Create ( SDL_Xft_Surface *xft_surface) {
    xft_surface->fcpattern=FcPatternCreate();
}

/* replace to  matched pattern */
void SDL_Xft_FontMatch(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface * xft_surface) {
    FcPattern *dup_fc ,*match; 
    FcResult result;

    dup_fc=FcPatternDuplicate(xft_surface->fcpattern);
    FcDefaultSubstitute(dup_fc);
    match=FcFontMatch (sdl_xft->config, dup_fc, &result);
    if (result != FcResultNoMatch) {
        FcPatternDestroy(xft_surface->fcpattern);
        xft_surface->fcpattern=FcPatternDuplicate(match);
    }
    FcPatternDestroy(match);
    FcPatternDestroy(dup_fc);
}

/* XftTextExtemts8 Wrapper */
Font_Rect SDL_Xft_TextExtents8 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, FcChar8 *msg, unsigned int length) {

    XftFont *xftfont=NULL;
    XftPattern *patterntmp=NULL;
    XGlyphInfo glyph;
    Font_Rect rect;

    Display *d;

    int s_len=0;

    d=sdl_xft->d;
    s_len=length;

    patterntmp=FcPatternDuplicate(xft_surface->fcpattern);

    if (s_len == 0) {
        while(msg[s_len] != (FcChar8)0) {
            s_len++;
        }
    }
    xftfont=XftFontOpenPattern (d, patterntmp);
    XftTextExtents8 (d, xftfont, msg, s_len, &glyph);
    XftFontClose(d, xftfont);

    rect.x=0;
    rect.y=0;
    rect.w=glyph.width;
    rect.h=glyph.height;
    rect.ascent=glyph.y;
    rect.descent=glyph.height-glyph.y;
    rect.woffset=glyph.xOff;

    return rect;
}

/* XftTextExtemts16 Wrapper */
Font_Rect SDL_Xft_TextExtents16 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, FcChar16 *msg, unsigned int length) {

    XftFont *xftfont=NULL;
    XftPattern *patterntmp=NULL;
    XGlyphInfo glyph;
    Font_Rect rect;

    Display *d;

    int s_len=0;

    d=sdl_xft->d;
    s_len=length;

    patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

    if (s_len == 0 )  {
        while(msg[s_len] != (FcChar16)0) {
            s_len++;
        }
    }
    xftfont=XftFontOpenPattern (d, patterntmp);

    XftTextExtents16 (d, xftfont, msg, s_len, &glyph);

    XftFontClose(d, xftfont);

    rect.x=0;
    rect.y=0;
    rect.w=glyph.width;
    rect.h=glyph.height;
    rect.ascent=glyph.y;
    rect.descent=glyph.height-glyph.y;
    rect.woffset=glyph.xOff;

    return rect;
}

/* XftTextExtemts32 Wrapper */
Font_Rect SDL_Xft_TextExtents32 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, FcChar32 *msg, unsigned int length) {

    XftFont *xftfont=NULL;
    XftPattern *patterntmp=NULL;
    XGlyphInfo glyph;
    Font_Rect rect;

    Display *d;

    int s_len=0;

    d=sdl_xft->d;
    s_len=length;

    patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

    if (s_len ==0) {
        while(msg[s_len] != (FcChar32)0) {
            s_len++;
        }
    }

    xftfont=XftFontOpenPattern (d, patterntmp);

    XftTextExtents32 (d, xftfont, msg, s_len, &glyph);

    XftFontClose(d, xftfont);

    rect.x=0;
    rect.y=0;
    rect.w=glyph.width;
    rect.h=glyph.height;
    rect.ascent=glyph.y;
    rect.descent=glyph.height-glyph.y;
    rect.woffset=glyph.xOff;

    return rect;
}

/* XftTextExtents32 Wrapper */
Font_Rect SDL_Xft_TextExtentsUtf8 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, FcChar8 *msg, unsigned int length) {

    XftFont *xftfont=NULL;
    XftPattern *patterntmp=NULL;
    XGlyphInfo glyph;
    Font_Rect rect;

    Display *d;

    int s_len=0;

    d=sdl_xft->d;
    s_len=length;

    patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

    if (s_len == 0) {
        while(msg[s_len] != (FcChar8)0) {
            s_len++;
        }
    }
    xftfont=XftFontOpenPattern (d, patterntmp);

    XftTextExtentsUtf8 (d, xftfont, msg, s_len, &glyph);

    XftFontClose(d, xftfont);

    rect.x=0;
    rect.y=0;
    rect.w=glyph.width;
    rect.h=glyph.height;
    rect.ascent=glyph.y;
    rect.descent=glyph.height-glyph.y;
    rect.woffset=glyph.xOff;

    return rect;
}

/* Draw Backcgound Pixmap */
void SDL_Xft_DrawRect (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, SDL_Rect *rect) {
    XRenderColor colortmp;
    XftColor bg;

    Display *d;

    d=sdl_xft->d;
    colortmp=xft_surface->color.bg;

    XftColorAllocValue(d, DefaultVisual(d, 0), 
            DefaultColormap(d,0), &colortmp, &bg);

    XftDrawRect(xft_surface->xftdraw, &bg, rect->x, rect->y, rect->w, rect->h);
}

/* draw 8bit string into pixmap of sdl_xft_surface  */
void SDL_Xft_DrawString8(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
        unsigned int x, unsigned int y, FcChar8 *msg) {
    XftFont *xftfont=NULL;
    XRenderColor colortmp;
    XftColor fg;
    XftPattern *patterntmp;

    Display *d;

    Font_Rect rect;
    int s_len=0;

    d=sdl_xft->d;
    colortmp=xft_surface->color.fg;

    patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

    while (msg[s_len] != (FcChar8)0) {
        s_len++;
    }

    xftfont=XftFontOpenPattern (d, patterntmp);

    XftColorAllocValue(d, DefaultVisual(d, 0), 
            DefaultColormap(d,0), &colortmp, &fg);

    if (xft_surface->xftdraw == NULL) {
        /* create xftdraw surface */
        rect=SDL_Xft_TextExtents8 (sdl_xft, xft_surface, msg, s_len);
        if (SDL_Xft_CreateSurface(sdl_xft, xft_surface, rect.w,  rect.h) !=0) {
            fprintf(stderr, "Can't create XftDraw surface !\n");
            return;
        }
    }

    XftDrawString8(xft_surface->xftdraw, &fg, xftfont, x, y,
            (FcChar8 * ) msg, s_len);
    XftFontClose(d, xftfont);
}

/* draw 16bit string into pixmap of sdl_xft_surface  */
void SDL_Xft_DrawString16(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
        unsigned int x, unsigned int y, FcChar16 *msg) {
    XftFont *xftfont=NULL;
    XRenderColor colortmp;
    XftColor fg;
    XftPattern *patterntmp;

    Display *d;

    Font_Rect rect;
    int s_len=0;

    d=sdl_xft->d;
    colortmp=xft_surface->color.fg;

    patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

    while (msg[s_len] != (FcChar16)0) {
        s_len++;
    }

    xftfont=XftFontOpenPattern (d, patterntmp);

    XftColorAllocValue(d, DefaultVisual(d, 0), 
            DefaultColormap(d,0), &colortmp, &fg);

    if (xft_surface->xftdraw == NULL) {
        /* create xftdraw surface */
        rect=SDL_Xft_TextExtents16 (sdl_xft, xft_surface, msg, s_len);
        if (SDL_Xft_CreateSurface(sdl_xft, xft_surface, rect.w,  rect.h) !=0) {
            fprintf(stderr, "Can't create XftDraw surface !\n");
            return;
        }
    }

    XftDrawString16(xft_surface->xftdraw, &fg, xftfont, x, y,
            (FcChar16 * ) msg, s_len);
    XftFontClose(d, xftfont);
}

/* draw 32bit string into pixmap of sdl_xft_surface  */
void SDL_Xft_DrawString32(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
        unsigned int x, unsigned int y, FcChar32 *msg) {
    XftFont *xftfont=NULL;
    XRenderColor colortmp;
    XftColor fg;
    XftPattern *patterntmp;

    Display *d;

    Font_Rect rect;
    int s_len=0;

    d=sdl_xft->d;
    colortmp=xft_surface->color.fg;

    patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

    while (msg[s_len] != (FcChar32)0) {
        s_len++;
    }

    xftfont=XftFontOpenPattern (d, patterntmp);

    XftColorAllocValue(d, DefaultVisual(d, 0), 
            DefaultColormap(d,0), &colortmp, &fg);

    if (xft_surface->xftdraw == NULL) {
        /* create xftdraw surface */
        rect=SDL_Xft_TextExtents32 (sdl_xft, xft_surface, msg, s_len);
        if (SDL_Xft_CreateSurface(sdl_xft, xft_surface, rect.w,  rect.h) !=0) {
            fprintf(stderr, "Can't create XftDraw surface !\n");
            return;
        }
    }

    XftDrawString32(xft_surface->xftdraw, &fg, xftfont, x, y,
            (FcChar32 * ) msg, s_len);
    XftFontClose(d, xftfont);
}

/* draw utf8 string into pixmap of sdl_xft_surface  */
void SDL_Xft_DrawStringUtf8(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
        unsigned int x, unsigned int y, FcChar8 *msg) {
    XftFont *xftfont=NULL;
    XRenderColor colortmp;
    XftColor fg;
    XftPattern *patterntmp;

    Display *d;

    Font_Rect rect;
    int s_len=0;

    d=sdl_xft->d;
    colortmp=xft_surface->color.fg;

    patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

    while (msg[s_len] != (FcChar8)0) {
        s_len++;
    }

    xftfont=XftFontOpenPattern (d, patterntmp);

    XftColorAllocValue(d, DefaultVisual(d, 0), 
            DefaultColormap(d,0), &colortmp, &fg);

    if (xft_surface->xftdraw == NULL) {
        /* create xftdraw surface */
        rect=SDL_Xft_TextExtentsUtf8 (sdl_xft, xft_surface, msg, s_len);
        if (SDL_Xft_CreateSurface(sdl_xft, xft_surface, rect.w,  rect.h) !=0) {
            fprintf(stderr, "Can't create XftDraw surface !\n");
            return;
        }
    }

    XftDrawStringUtf8(xft_surface->xftdraw, &fg, xftfont, x, y,
            (FcChar8 * ) msg, s_len);
    XftFontClose(d, xftfont);
}

/*fix me */
/* bliting pixmap into SDL_Surface */
int SDL_Xft_DrawToSDLSurface(SDL_Xft_WMInfo *sdl_xft, SDL_Surface *dest, SDL_Xft_Surface *src,
        SDL_Rect *destrect, SDL_Rect *srcrect) {
    Display *d;
    XImage *ximg;

    XftColor ckey;
    XRenderColor colortmp;

    SDL_Rect drawrect;

    unsigned int i=0, j=0;
    unsigned long pixel;
    unsigned int key;
    unsigned int dest_x=0, dest_y=0;
    Uint32 mpixel;

    if (dest == NULL || src == NULL) {
        return 0;
    }

    if (srcrect != NULL) {
        drawrect.x=srcrect->x;
        drawrect.y=srcrect->y;
        drawrect.w=srcrect->w;
        drawrect.h=srcrect->h;
    }
    else {
        drawrect.x=0;
        drawrect.y=0;
        drawrect.w=src->width;
        drawrect.h=src->height;
    }

    d=sdl_xft->d;
    key=src->color.colorkey;
    switch (key) {
        case SDL_XFT_NO_COLORKEY :
            break;
        case SDL_XFT_BG_COLORKEY :
            colortmp=src->color.bg;
            break;
        case SDL_XFT_FG_COLORKEY :
            colortmp=src->color.fg;
            break;
        default :
            break;
    }

    if (key > 0) {
        XftColorAllocValue(d, DefaultVisual(d, 0), DefaultColormap(d,0), &colortmp, &ckey);
    }

    /*
       GetDrawRect(&drawrect, &destrect, &srcrect);
       */

    if (destrect != NULL )  {
        dest_x=destrect->x;
        dest_y=destrect->y;
    }

    ximg=XGetImage(sdl_xft->d, src->pixmap, 
            /*
               drawrect.x, drawrect.y, drawrect.w, drawrect.h, -1, ZPixmap);
               */
        0, 0, src->width, src->height, -1, ZPixmap);


    for ( i=drawrect.y ; i < drawrect.h ; i++) {
        for (j=drawrect.x; j< drawrect.w ; j++) {
            pixel=XGetPixel(ximg, j,i);
            if (key > 0  && pixel == ckey.pixel) {
                continue;
            }

            mpixel=GetMappedPixel(sdl_xft, dest, pixel);
            putpixel(dest, j, i, mpixel);
            dest_x++;
            dest_y++;
            /* for debug
               printf("%#lx\t", pixel);
               fprintf(stdout, "%d\t%d\t%d\t%d\n", i,j, drawrect.h, drawrect.w);
               */
        }
    }
    return 1;
}

SDL_Surface *SDL_Xft_GetSurface(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface,
        int width, int height, SDL_Rect *srcrect) {
    SDL_Surface *scr;
    SDL_Rect rect;
    /* create SDL_Surface */
    int surface_w=0;
    int surface_h=0;

    /*initialize values */
    if (srcrect == NULL) {
        rect.x=0;
        rect.y=0;
        rect.w=xft_surface->width;
        rect.h=xft_surface->height;
    }
    else {
        rect.x=srcrect->x;
        rect.y=srcrect->y;
        rect.w=srcrect->w;
        rect.h=srcrect->h;
    }

    if (width == 0 ) {
        surface_w=xft_surface->width;
    }
    else {
        surface_w=width;
    }

    if (height == 0 ) {
        surface_h=xft_surface->height;
    }
    else {
        surface_w=height;
    }

    /* Create Surface */
    scr=SDL_CreateRGBSurface(xft_surface->flag , surface_w,  surface_h, 
            sdl_xft->screen->format->BitsPerPixel, 
            sdl_xft->screen->format->Rmask, 
            sdl_xft->screen->format->Gmask, 
            sdl_xft->screen->format->Bmask, 
            sdl_xft->screen->format->Amask);
    if (scr == NULL ) {
        return NULL;
    }
    if (SDL_Xft_DrawToSDLSurface (sdl_xft, scr, xft_surface , NULL, &rect) == 0) {
        return NULL;
    }

    return scr;
}


void SDL_Xft_Quit(SDL_Xft_WMInfo *sdl_xft) {
    FcConfigDestroy(sdl_xft->config);
}


void SDL_Xft_Surface_Free (SDL_Xft_Surface *xft_surface){
    if (xft_surface->fcpattern != NULL) {
        FcPatternDestroy(xft_surface->fcpattern);
    }
    if (xft_surface->xftdraw != NULL) {
        XftDrawDestroy(xft_surface->xftdraw);
    }
    free(xft_surface);
}

/* private function */
unsigned int FcPattern_Exist(FcPattern *fcpattern) {
    if (fcpattern == NULL) {
        return 0;
    }
    else {
        return 1;
    }
}


void GetDrawRect (SDL_Rect *draw, SDL_Rect *dest, SDL_Rect * src) {
    int dest_x2=0, src_x2=0;
    int dest_y2=0, src_y2=0;

    /* Temp x2, y2 axis */
    dest_x2=dest->x + dest->w;
    src_x2=src->x + dest->w;
    dest_y2=dest->y + dest->h;
    src_y2=src->x + src->h;

    /* get X, y axis */
    if (dest->x < src->x) {
        draw->x=dest->x;
    }
    else {
        draw->x=src->x;
    }

    if (dest->y < src->y) {
        draw->y=dest->y;
    }

    /* get w, h */
    if (dest_x2 < src_x2) {
        draw->w=(dest_x2-draw->x);
    }
    else {
        draw->w=(src_x2-draw->x);
    }

    if (dest_y2 < src_y2) {
        draw->h= (dest_y2- draw->y);
    }
    else {
        draw->h= (src_y2- draw->y);
    }

}
