#ifndef _SDL_XFT_
#define _SDL_XFT_

#include <X11/Xlib.h>
#include <X11/Xft/Xft.h>
#include <stdarg.h>
#include <fontconfig/fontconfig.h>

#include "SDL.h"
#include "SDL_syswm.h"

#define DEFAULT 0
#define COMPACT 1


#define PIXMAP_ERROR 1
#define XFTDRAW_ERROR 2

typedef struct _Font_Rect {
    unsigned int x;
    unsigned int y; 
    unsigned int w; 
    unsigned int h;
    unsigned int ascent;
    unsigned int descent;
    unsigned int woffset;
}Font_Rect;

typedef struct _Xft_RGB_Format {
    int r_rshift;
    int g_rshift;
    int b_rshift;
    int r_lshift;
    int g_lshift;
    int b_lshift;
}Xft_RGB_Format;

/* SDL Xft information */
typedef struct _SDL_Xft_WMInfo {
    Display *d;
    Window w;
    Visual *visual;
    SDL_Surface *screen;
    FcConfig *config;
    Xft_RGB_Format format;
}SDL_Xft_WMInfo;

/* color key definitaion */
#define SDL_XFT_NO_COLORKEY 	0
#define SDL_XFT_BG_COLORKEY 	1
#define SDL_XFT_FG_COLORKEY 	2

/* color information */
typedef struct _SDL_Xft_Color {
    XRenderColor fg;
    XRenderColor bg;
    unsigned int colorkey;
}SDL_Xft_Color;

/* SDL Xft data struct */
typedef struct _SDL_Xft_Surface {
    Pixmap pixmap;
    SDL_Xft_Color color;
    FcPattern *fcpattern;
    XftDraw *xftdraw;
    Uint32 flag;		/* SDL_Surface flag */
    unsigned int width;	/* pixmap width*/
    unsigned int height;	/* pixmap height */
    int compact;		/* when compact flag set, pixmap createed by string size*/
    Font_Rect used;		/* used font rectangle area */

}SDL_Xft_Surface;


/* init SDL_Xft_WmInfo */
int SDL_Xft_Init ( SDL_Xft_WMInfo *sdl_xft , SDL_Surface *screen);

/* get system font set, return number of font list */
FcChar8 ** SDL_Xft_Get_FontList (int *fnum, SDL_Xft_WMInfo *sdl_xft);

SDL_Xft_Surface * SDL_Xft_Create(Uint32 flag,  SDL_Xft_WMInfo *sdl_xft);

/* Create pixmap xftdraw sturcture */
int SDL_Xft_CreateSurface(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *surface, unsigned int width, unsigned int height);

SDL_Xft_Surface * SDL_Xft_Create_Surface_Compact (Uint32 flag, SDL_Xft_WMInfo *sdl_xft);
/* set allocation color */
#define SDL_XFT_BG_COLOR 0
#define SDL_XFT_FG_COLOR 1

void SDL_Xft_Color_Alloc ( SDL_Xft_Color *color , unsigned int flag, 
        Uint16 r, Uint16 g, Uint16 b);

/* set colorkey */
void SDL_Xft_Set_Colorkey (SDL_Xft_Color *color , unsigned int flag);

/* FcPatternAddInteger Wrapper */
int SDL_Xft_Pattern_AddInteger (SDL_Xft_Surface *xft_surface,
        const char *object, int i);

/* FcPatternAddDoubel Wrapper */
int SDL_Xft_Pattern_AddDouble (SDL_Xft_Surface *xft_surface, 
        const char *object, double d);

/* FcPatternAddString Wrapper */
int SDL_Xft_Pattern_AddString (SDL_Xft_Surface *xft_surface, 
        const char *object, const char *s);

/* FcPatternAddMatrix Wrapper */
int SDL_Xft_Pattern_AddMatrix (SDL_Xft_Surface *xft_surface,
        const char *object, const FcMatrix *s);

/* FcPatternAddCharSet Wrapper */
int SDL_Xft_Pattern_AddCharSet (SDL_Xft_Surface *xft_surface,
        const char *object, const FcCharSet *c);

/* FcPatternAddBool Wrapper */
int SDL_Xft_Pattern_AddBool (SDL_Xft_Surface *xft_surface,
        const char *object, FcBool b);

#define FC_PATTERN_GET_FAIL 0
#define FC_PATTERN_GET_SUCCESS 1

/* FcPatternGetInteger Wrapper */
unsigned int SDL_Xft_Pattern_GetInteger(SDL_Xft_Surface *xft_surface, const char *object, int *i);

/* FcPatternGetDouble Wrapper */
unsigned int SDL_Xft_Pattern_GetDouble(SDL_Xft_Surface *xft_surface, const char *object, double *d);

/* FcPatternGetString Wrapper */
unsigned int SDL_Xft_Pattern_GetString(SDL_Xft_Surface *xft_surface, const char *object, FcChar8 **consts);

/* FcPatternGetMatrix Wrapper */
unsigned int SDL_Xft_Pattern_GetMatrix(SDL_Xft_Surface *xft_surface, const char *object, FcMatrix **s);

/* FcPatternGetCharSet Wrapper */
unsigned int SDL_Xft_Pattern_GetCharSet(SDL_Xft_Surface *xft_surface, const char *object, FcCharSet **c);

/* FcPatternGetBool Wrapper */
unsigned int SDL_Xft_Pattern_GetBool(SDL_Xft_Surface *xft_surface, const char *object, FcBool *b);

/* FcPatternBuild Wrapper */
void SDL_Xft_Pattern_Build (SDL_Xft_Surface *xft_surface, ...);

/* Create pattern as default */
void SDL_Xft_Pattern_Create(SDL_Xft_Surface *xft_surface);

/* replace to  matched pattern */
void SDL_Xft_FontMatch(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface);

/* XftTextExtemts8 Wrapper */
Font_Rect SDL_Xft_TextExtents8 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, FcChar8 *msg, unsigned int length);

/* XftTextExtemts16 Wrapper */
Font_Rect SDL_Xft_TextExtents16 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, FcChar16 *msg, unsigned int length);

/* XftTextExtemts32 Wrapper */
Font_Rect SDL_Xft_TextExtents32 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, FcChar32 *msg, unsigned int length);

/* XftTextExtemtsUtf8 Wrapper */
Font_Rect SDL_Xft_TextExtentsUtf8 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, FcChar8 *msg, unsigned int length);

/* Draw Background into SDL_Xft_Surface->Pixmap */
void SDL_Xft_DrawRect (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, SDL_Rect *rect);

/* Draw 8bit string into SDL_Xft_Surface->Pixmap */
void SDL_Xft_DrawString8(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
        unsigned int x, unsigned int y, FcChar8 *msg);

/* Draw 16bit string into SDL_Xft_Surface->Pixmap */
void SDL_Xft_DrawString16(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
        unsigned int x, unsigned int y, FcChar16 *msg);

/* Draw 32bit string into SDL_Xft_Surface->Pixmap */
void SDL_Xft_DrawString32(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
        unsigned int x, unsigned int y, FcChar32 *msg);

/* Draw utf8 string into SDL_Xft_Surface->Pixmap */
void SDL_Xft_DrawStringUtf8(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
        unsigned int x, unsigned int y, FcChar8 *msg);
/* flag may be SDL_Xft_String type */




/* Create Font blited surface */
SDL_Surface *SDL_Xft_GetSurface(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface,
        int widtht, int heigh, SDL_Rect *srcrect);

/*list of free function */
/* free SDL_Xft data struct */
void SDL_Xft_Quit(SDL_Xft_WMInfo *sdl_xft);

void SDL_Xft_Surface_Free (SDL_Xft_Surface *xft_surface);

void SDL_Xft_FontList_Free (FcChar8 **f_list);

/* private function */
unsigned int FcPattern_Exist(FcPattern *fcpattern); 

/* Bliting pixmap to surface */
int SDL_Xft_DrawToSDLSurface(SDL_Xft_WMInfo *sdl_xft, SDL_Surface *dest, SDL_Xft_Surface *src,
        SDL_Rect *destrect, SDL_Rect *srcrect);

void GetDrawRect (SDL_Rect *draw, SDL_Rect *dest, SDL_Rect * src);

/*Create XftDraw sturct Function */

#endif /* _SDL_XFT_ */
